@@ -24,6 +24,7 @@
 #include <locale.h>
 #include <errno.h>
 #include <sys/types.h>
+#include <sys/resource.h>
 #include <unistd.h>
 
 #ifdef HAVE_MALLOC_TRIM
@@ -35,11 +36,15 @@
 #include <glib-object.h>
 #include <glib/gi18n.h>
 
+#include <gst/gst.h>
+
 #include <libtracker-miners-common/tracker-common.h>
 #include <libtracker-sparql/tracker-sparql.h>
 
 #include "tracker-config.h"
+#include "tracker-controller.h"
 #include "tracker-miner-files.h"
+#include "tracker-files-interface.h"
 
 #define ABOUT	  \
 	"Tracker " PACKAGE_VERSION "\n"
@@ -55,11 +60,8 @@
 
 #define DBUS_NAME_SUFFIX "Tracker3.Miner.Files"
 #define DBUS_PATH "/org/freedesktop/Tracker3/Miner/Files"
-#define LOCALE_FILENAME "locale-for-miner-apps.txt"
 
 static GMainLoop *main_loop;
-static GDBusProxy *index_proxy;
-static GPtrArray *proxy_folders;
 static guint cleanup_id;
 
 static gint initial_sleep = -1;
@@ -142,33 +144,6 @@ log_option_values (TrackerConfig *config)
 #endif
 }
 
-static void
-miner_reset_applications (TrackerMiner *miner)
-{
-	const gchar *sparql;
-	GError *error = NULL;
-
-	sparql =
-		"DELETE { ?icon a rdfs:Resource } "
-		"WHERE { ?app a nfo:SoftwareApplication; nfo:softwareIcon ?icon }; "
-		"DELETE { ?app tracker:extractorHash ?h } "
-		"WHERE { ?app a nfo:SoftwareApplication ; "
-		"             tracker:extractorHash ?h } ";
-
-	/* Execute a sync update, we don't want the apps miner to start before
-	 * we finish this. */
-	tracker_sparql_connection_update (tracker_miner_get_connection (miner),
-	                                  sparql,
-	                                  NULL, &error);
-
-	if (error) {
-		/* Some error happened performing the query, not good */
-		g_critical ("Couldn't reset indexed applications: %s",
-		            error ? error->message : "unknown error");
-		g_error_free (error);
-	}
-}
-
 static GFile *
 get_cache_dir (TrackerDomainOntology *domain_ontology)
 {
@@ -178,99 +153,6 @@ get_cache_dir (TrackerDomainOntology *domain_ontology)
 	return g_file_get_child (cache, "files");
 }
 
-static void
-save_current_locale (TrackerDomainOntology *domain_ontology)
-{
-	GError *error = NULL;
-	gchar *locale = tracker_locale_get (TRACKER_LOCALE_LANGUAGE);
-	GFile *cache = get_cache_dir (domain_ontology);
-	gchar *cache_path = g_file_get_path (cache);
-	gchar *locale_file;
-
-	locale_file = g_build_filename (cache_path, LOCALE_FILENAME, NULL);
-	g_free (cache_path);
-
-	TRACKER_NOTE (CONFIG, g_message ("Saving locale used to index applications"));
-	TRACKER_NOTE (CONFIG, g_message ("  Creating locale file '%s'", locale_file));
-
-	if (locale == NULL) {
-		locale = g_strdup ("");
-	}
-
-	if (!g_file_set_contents (locale_file, locale, -1, &error)) {
-		g_message ("  Could not set file contents, %s",
-		           error ? error->message : "no error given");
-		g_clear_error (&error);
-	}
-
-	g_object_unref (cache);
-	g_free (locale);
-	g_free (locale_file);
-}
-
-static gboolean
-detect_locale_changed (TrackerMiner          *miner,
-		       TrackerDomainOntology *domain_ontology)
-{
-	GFile *cache;
-	gchar *cache_path;
-	gchar *locale_file;
-	gchar *previous_locale = NULL;
-	gchar *current_locale;
-	gboolean changed;
-
-	cache = get_cache_dir (domain_ontology);
-	cache_path = g_file_get_path (cache);
-	locale_file = g_build_filename (cache_path, LOCALE_FILENAME, NULL);
-	g_object_unref (cache);
-	g_free (cache_path);
-
-	if (G_LIKELY (g_file_test (locale_file, G_FILE_TEST_EXISTS))) {
-		gchar *contents;
-
-		/* Check locale is correct */
-		if (G_LIKELY (g_file_get_contents (locale_file, &contents, NULL, NULL))) {
-			if (contents &&
-			    contents[0] == '\0') {
-				g_critical ("  Empty locale file found at '%s'", locale_file);
-				g_free (contents);
-			} else {
-				/* Re-use contents */
-				previous_locale = contents;
-			}
-		} else {
-			g_critical ("  Could not get content of file '%s'", locale_file);
-		}
-	} else {
-		TRACKER_NOTE (CONFIG, g_message ("  Could not find locale file:'%s'", locale_file));
-	}
-
-	g_free (locale_file);
-
-	current_locale = tracker_locale_get (TRACKER_LOCALE_LANGUAGE);
-
-	/* Note that having both to NULL is actually valid, they would default
-	 * to the unicode collation without locale-specific stuff. */
-	if (g_strcmp0 (previous_locale, current_locale) != 0) {
-		TRACKER_NOTE (CONFIG, g_message ("Locale change detected from '%s' to '%s'...",
-		                      previous_locale, current_locale));
-		changed = TRUE;
-	} else {
-		TRACKER_NOTE (CONFIG, g_message ("Current and previous locales match: '%s'", previous_locale));
-		changed = FALSE;
-	}
-
-	g_free (current_locale);
-	g_free (previous_locale);
-
-	if (changed) {
-		TRACKER_NOTE (CONFIG, g_message ("Resetting nfo:Software due to locale change..."));
-		miner_reset_applications (miner);
-	}
-
-	return changed;
-}
-
 static gboolean
 signal_handler (gpointer user_data)
 {
@@ -339,6 +221,19 @@ initialize_priority_and_scheduling (void)
 	}
 }
 
+static void
+raise_file_descriptor_limit (void)
+{
+	struct rlimit rl;
+
+	if (getrlimit (RLIMIT_NOFILE, &rl) != 0)
+		return;
+
+	rl.rlim_cur = rl.rlim_max;
+	if (setrlimit(RLIMIT_NOFILE, &rl) != 0)
+		g_warning ("Failed to increase file descriptor limit: %m");
+}
+
 static gboolean
 should_crawl (TrackerMinerFiles *miner_files,
               TrackerConfig     *config,
@@ -529,43 +424,6 @@ miner_finished_cb (TrackerMinerFS *fs,
 	}
 }
 
-static GList *
-get_dir_children_as_gfiles (const gchar *path)
-{
-	GList *children = NULL;
-	GDir *dir;
-
-	dir = g_dir_open (path, 0, NULL);
-
-	if (dir) {
-		const gchar *basename;
-
-		while ((basename = g_dir_read_name (dir)) != NULL) {
-			GFile *child;
-			gchar *str;
-
-			str = g_build_filename (path, basename, NULL);
-			child = g_file_new_for_path (str);
-			g_free (str);
-
-			children = g_list_prepend (children, child);
-		}
-
-		g_dir_close (dir);
-	}
-
-	return children;
-}
-
-static void
-dummy_log_handler (const gchar    *domain,
-                   GLogLevelFlags  log_level,
-                   const gchar    *message,
-                   gpointer        user_data)
-{
-	return;
-}
-
 static void
 graphs_created_cb (GObject      *source,
                    GAsyncResult *res,
@@ -580,32 +438,20 @@ graphs_created_cb (GObject      *source,
 }
 
 static gint
-check_eligible (void)
+check_eligible (TrackerIndexingTree *indexing_tree,
+                TrackerStorage      *storage)
 {
-	TrackerSparqlConnection *sparql_conn;
-	TrackerMiner *miner_files;
-	TrackerIndexingTree *indexing_tree;
-	TrackerDomainOntology *domain_ontology;
-	TrackerConfig *config;
-	GFile *ontology;
+	TrackerController *controller;
 	GFile *file;
 	GFileInfo *info;
 	GError *error = NULL;
 	gchar *path;
-	guint log_handler_id;
 	gboolean exists = TRUE;
 	gboolean indexable;
 	gboolean parents_indexable = TRUE;
 	gboolean is_dir;
 
-	/* Set log handler for library messages */
-	log_handler_id = g_log_set_handler (NULL,
-	                                    G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL,
-	                                    dummy_log_handler,
-	                                    NULL);
-
-	g_log_set_default_handler (dummy_log_handler, NULL);
-
+	controller = tracker_controller_new (indexing_tree, storage, NULL);
 
 	/* Start check */
 	file = g_file_new_for_commandline_arg (eligible);
@@ -631,7 +477,6 @@ check_eligible (void)
 		is_dir = FALSE;
 	}
 
-	config = tracker_config_new ();
 	path = g_file_get_path (file);
 
 	g_print (exists ?
@@ -641,34 +486,6 @@ check_eligible (void)
 
 	g_print ("\n");
 
-	domain_ontology = tracker_domain_ontology_new (domain_ontology_name, NULL, &error);
-	ontology = tracker_domain_ontology_get_ontology (domain_ontology);
-	tracker_domain_ontology_unref (domain_ontology);
-
-	sparql_conn = tracker_sparql_connection_new (0,
-	                                              NULL,
-	                                              ontology,
-	                                              NULL,
-	                                              NULL);
-	if (!sparql_conn) {
-		g_object_unref (info);
-		return EXIT_FAILURE;
-	}
-
-	/* Create new TrackerMinerFiles object */
-	config = tracker_config_new ();
-	miner_files = tracker_miner_files_new (sparql_conn, config,
-	                                       domain_ontology_name, NULL);
-	g_object_unref (config);
-
-	if (!miner_files) {
-		g_object_unref (info);
-		g_object_unref (sparql_conn);
-		return EXIT_FAILURE;
-	}
-
-	indexing_tree = tracker_miner_fs_get_indexing_tree (TRACKER_MINER_FS (miner_files));
-
 	indexable = tracker_indexing_tree_file_is_indexable (indexing_tree, file, info);
 
 	if (!indexable) {
@@ -718,16 +535,12 @@ check_eligible (void)
 				g_print ("\n");
 				parents_indexable = FALSE;
 			} else {
-				GList *children = get_dir_children_as_gfiles (dir_path);
-
-				if (!tracker_indexing_tree_parent_is_indexable (indexing_tree, l->data, children)) {
+				if (!tracker_indexing_tree_parent_is_indexable (indexing_tree, l->data)) {
 					g_print (_("Parent directory “%s” is NOT eligible to be indexed (based on content filters)"),
 					         dir_path);
 					g_print ("\n");
 					parents_indexable = FALSE;
 				}
-
-				g_list_free_full (children, g_object_unref);
 			}
 
 			g_free (dir_path);
@@ -746,15 +559,10 @@ check_eligible (void)
 		         _("File is eligible to be indexed"));
 	}
 
-	if (log_handler_id != 0) {
-		/* Unset log handler */
-		g_log_remove_handler (NULL, log_handler_id);
-	}
-
 	g_free (path);
 	g_object_unref (file);
-	g_object_unref (miner_files);
 	g_object_unref (info);
+	g_object_unref (controller);
 
 	return (indexable && parents_indexable) ? EXIT_SUCCESS : EXIT_FAILURE;
 }
@@ -914,126 +722,18 @@ setup_connection (TrackerDomainOntology  *domain,
 
 	if (!dry_run)
 		store = get_cache_dir (domain);
-	ontology = tracker_domain_ontology_get_ontology (domain);
+	ontology = tracker_sparql_get_ontology_nepomuk ();
 	sparql_conn = tracker_sparql_connection_new (get_fts_connection_flags (),
 	                                             store,
 	                                             ontology,
 	                                             NULL,
 	                                             error);
 	g_clear_object (&store);
+	g_clear_object (&ontology);
 
 	return sparql_conn;
 }
 
-static void
-update_indexed_files_from_proxy (TrackerMinerFiles *miner,
-                                 GDBusProxy        *proxy)
-{
-	TrackerIndexingTree *indexing_tree;
-	const gchar **indexed_uris = NULL;
-	GVariant *v;
-	gint i;
-
-	v = g_dbus_proxy_get_cached_property (proxy, "IndexedLocations");
-	if (v)
-		indexed_uris = g_variant_get_strv (v, NULL);
-
-	indexing_tree = tracker_miner_fs_get_indexing_tree (TRACKER_MINER_FS (miner));
-
-	/* Remove folders no longer there */
-	for (i = 0; i < proxy_folders->len; i++) {
-		GFile *file;
-		gchar *uri;
-
-		file = g_ptr_array_index (proxy_folders, i);
-		uri = g_file_get_uri (file);
-
-		if (!indexed_uris || !g_strv_contains (indexed_uris, uri)) {
-			tracker_indexing_tree_remove (indexing_tree,
-			                              file);
-		}
-
-		g_free (uri);
-	}
-
-	for (i = 0; indexed_uris && indexed_uris[i]; i++) {
-		GFileInfo *file_info;
-		GFile *file;
-
-		file = g_file_new_for_uri (indexed_uris[i]);
-		if (g_ptr_array_find_with_equal_func (proxy_folders,
-		                                      file,
-		                                      (GEqualFunc) g_file_equal,
-		                                      NULL)) {
-			g_object_unref (file);
-			continue;
-		}
-
-		file_info = g_file_query_info (file,
-		                               G_FILE_ATTRIBUTE_STANDARD_TYPE ","
-		                               G_FILE_ATTRIBUTE_STANDARD_IS_HIDDEN,
-		                               G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
-		                               NULL, NULL);
-
-		if (!file_info) {
-			g_object_unref (file);
-			continue;
-		}
-
-		if (g_file_info_get_file_type (file_info) == G_FILE_TYPE_DIRECTORY) {
-			if (!tracker_indexing_tree_file_is_indexable (indexing_tree,
-			                                              file, file_info)) {
-				tracker_indexing_tree_add (indexing_tree,
-				                           file,
-				                           TRACKER_DIRECTORY_FLAG_RECURSE |
-				                           TRACKER_DIRECTORY_FLAG_CHECK_MTIME |
-				                           TRACKER_DIRECTORY_FLAG_MONITOR);
-				g_ptr_array_add (proxy_folders, g_object_ref (file));
-			} else {
-				tracker_indexing_tree_notify_update (indexing_tree,
-				                                     file, TRUE);
-			}
-		} else {
-			tracker_miner_fs_check_file (TRACKER_MINER_FS (miner),
-			                             file, G_PRIORITY_HIGH, TRUE);
-		}
-
-		g_object_unref (file_info);
-		g_object_unref (file);
-	}
-
-	g_free (indexed_uris);
-}
-
-static void
-proxy_properties_changed_cb (GDBusProxy *proxy,
-                             GVariant   *changed_properties,
-                             GStrv       invalidated_properties,
-                             gpointer    user_data)
-{
-	update_indexed_files_from_proxy (user_data, proxy);
-}
-
-static void
-on_tracker_index_proxy_ready (GObject      *source,
-                              GAsyncResult *res,
-                              gpointer      user_data)
-{
-	TrackerMinerFiles *miner = user_data;
-	GError *error = NULL;
-
-	index_proxy = g_dbus_proxy_new_finish (res, &error);
-	if (error) {
-		g_critical ("Could not set up proxy: %s", error->message);
-		g_error_free (error);
-		return;
-	}
-
-	g_signal_connect (index_proxy, "g-properties-changed",
-	                  G_CALLBACK (proxy_properties_changed_cb), miner);
-	update_indexed_files_from_proxy (miner, index_proxy);
-}
-
 int
 main (gint argc, gchar *argv[])
 {
@@ -1046,12 +746,15 @@ main (gint argc, gchar *argv[])
 	TrackerMinerProxy *proxy;
 	GDBusConnection *connection;
 	TrackerSparqlConnection *sparql_conn;
+	TrackerIndexingTree *indexing_tree;
+	TrackerStorage *storage;
 	TrackerDomainOntology *domain_ontology;
-	GCancellable *cancellable;
+	TrackerController *controller;
 #if GLIB_CHECK_VERSION (2, 64, 0)
 	GMemoryMonitor *memory_monitor;
 #endif
 	gchar *domain_name, *dbus_name;
+	TrackerFilesInterface *files_interface;
 
 	main_loop = NULL;
 
@@ -1067,6 +770,16 @@ main (gint argc, gchar *argv[])
 	/* This makes sure we don't steal all the system's resources */
 	initialize_priority_and_scheduling ();
 
+	/* This makes it harder to run out of file descriptors while there
+	 * are many concurrently running queries through the endpoint.
+	 */
+	raise_file_descriptor_limit ();
+
+	/* Preempt possible registry updates, before tracker-extract-3 deals
+	 * with gstreamer plugins.
+	 */
+	gst_init (NULL, NULL);
+
 	/* Translators: this messagge will apper immediately after the
 	 * usage string - Usage: COMMAND <THIS_MESSAGE>
 	 */
@@ -1087,8 +800,11 @@ main (gint argc, gchar *argv[])
 		return EXIT_SUCCESS;
 	}
 
+	indexing_tree = tracker_indexing_tree_new ();
+	storage = tracker_storage_new ();
+
 	if (eligible) {
-		return check_eligible ();
+		return check_eligible (indexing_tree, storage);
 	}
 
 	domain_ontology = tracker_domain_ontology_new (domain_ontology_name, NULL, &error);
@@ -1107,6 +823,8 @@ main (gint argc, gchar *argv[])
 		return EXIT_FAILURE;
 	}
 
+	files_interface = tracker_files_interface_new (connection);
+
 	/* Initialize logging */
 	config = tracker_config_new ();
 
@@ -1161,29 +879,13 @@ main (gint argc, gchar *argv[])
 	}
 
 	/* Create new TrackerMinerFiles object */
-	miner_files = tracker_miner_files_new (sparql_conn, config,
-	                                       domain_ontology_name, &error);
-	if (!miner_files) {
-		g_critical ("Couldn't create new Files miner: '%s'",
-		            error ? error->message : "unknown error");
-		g_object_unref (config);
-		return EXIT_FAILURE;
-	}
+	miner_files = tracker_miner_files_new (sparql_conn,
+	                                       indexing_tree,
+	                                       storage,
+	                                       config,
+	                                       domain_ontology);
 
-	proxy_folders = g_ptr_array_new_with_free_func (g_object_unref);
-	cancellable = g_cancellable_new ();
-	g_dbus_proxy_new (connection,
-	                  G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
-	                  NULL,
-	                  "org.freedesktop.Tracker3.Miner.Files.Control",
-	                  "/org/freedesktop/Tracker3/Miner/Files/Proxy",
-	                  "org.freedesktop.Tracker3.Miner.Files.Proxy",
-	                  cancellable,
-	                  on_tracker_index_proxy_ready,
-	                  miner_files);
-
-	/* If the locales changed, we need to reset some things first */
-	detect_locale_changed (TRACKER_MINER (miner_files), domain_ontology);
+	controller = tracker_controller_new (indexing_tree, storage, files_interface);
 
 	proxy = tracker_miner_proxy_new (miner_files, connection, DBUS_PATH, NULL, &error);
 	if (error) {
@@ -1273,20 +975,17 @@ main (gint argc, gchar *argv[])
 
 	g_debug ("Shutdown started");
 
-	if (!dry_run && miners_timeout_id == 0 && !miner_needs_check (miner_files)) {
+	if (!dry_run && miners_timeout_id == 0 && !miner_needs_check (miner_files))
 		tracker_miner_files_set_need_mtime_check (TRACKER_MINER_FILES (miner_files), FALSE);
-		save_current_locale (domain_ontology);
-	}
 
 	finish_endpoint_thread ();
 
+	g_object_unref (files_interface);
+
 	g_main_loop_unref (main_loop);
 	g_object_unref (config);
 
-	g_cancellable_cancel (cancellable);
-	g_object_unref (cancellable);
-	g_clear_object (&index_proxy);
-	g_clear_pointer (&proxy_folders, g_ptr_array_unref);
+	g_object_unref (controller);
 
 	g_object_unref (miner_files);
 
@@ -1297,6 +996,9 @@ main (gint argc, gchar *argv[])
 	tracker_sparql_connection_close (sparql_conn);
 	g_object_unref (sparql_conn);
 
+	g_clear_object (&indexing_tree);
+	g_clear_object (&storage);
+
 #if GLIB_CHECK_VERSION (2, 64, 0)
 	g_signal_handlers_disconnect_by_func (memory_monitor, on_low_memory, NULL);
 	g_object_unref (memory_monitor);
